************************************************************
**
** RPL OBJECT 
**
** Name: InterStack - The interactive stack application
**
** Category: User Interface
**
** File: InterStack.s
**
** Purpose:
**   Provides an interface to select and operate on stack items.
**
** Stack on Entry: -
**   
** Other Entry Conditions:
**
** Stack on Exit:  -
**
** Error Exits:
**
** Detail:
**   The interactive stack maintains three variables defining the
**   selected stack level:
**
**     SelLev	(selected level) The stack level number of 
**              the selected level, from #1 to DEPTH
**     SelPos	(selected position) The display row number
**              of the selected level, from #1 (top) to 
**              #4 (bottom).
**     SelLevDelta (selected level offset) The difference between
**              the true selected stack level number and the 
**              user stack level number.  System variable 
**              "DEPTHSAVE" is used to compute this.
**
**********************************************************************

** Macro definitions:
DEFINE ISExit!	    1PUTLAM
DEFINE ISExit@	    1GETLAM
DEFINE SelPos!      2PUTLAM
DEFINE SelPos@      2GETLAM
DEFINE SelLev!      3PUTLAM
DEFINE SelLev@      3GETLAM
DEFINE SelLevDelta! 4PUTLAM
DEFINE SelLevDelta@ 4GETLAM

NULLNAME IStk		( --> )
::
  TakeOver

  UStackDepth 			( get the user stack depth )
  #0=case 			( if the user stack is empty... )
    ::  			( ...then do a "Stack Empty" warning )
      EmptyStack$ DoWarning
      SetDA23NoCh 
    ;

  NoEditLine? IT 	( if there's no edit line... )
    ::			( ...then perform important aspects of "StdEnter" )
      CacheStack	( if last stack mode is enabled then save the stack )
      InitEdModes	( reset the edit modes )
    ;


  POLSaveUI		( save the current user interface )
  ERRSET 		( prepare to restore saved user interface in
			case of error )
    ::
      DEPTH UStackDepth #- 	( calculate the selected level offset )
      ONE FIVE 		( select stack level 1 and display position 5 )

      FALSE		( initialize exit condition to FALSE )

      ' NULLLAM FOUR NDUPN DOBIND	( save IStk variables )
      ' IStkDisp	( specify IStk display object )
      ' IStkKeys	( specify IStk key handler )
      TrueTrue		( specify IStk and standard
			key handling )
      EditLExists? 	( if there's an edit line... )
      SelLevDelta@ #0<> OR ITE
	( ...OR the selected level offset is non-zero... )
        :: RESOROMP IStkMenuEd	; ( ...then specify the editing IStk menu )
        :: RESOROMP IStkMenu ;	( ...else specify the non-editing IStk menu )
      ONEFALSE		( specify the IStk menu row and
  			no suspended environments allowed )
      ' ISExit@		( specify IStk exit condition )
      'ERRJMP		( specify IStk error handler )
      POLSetUI		( set the IStk user interface )

      ClrDAsOK		( insure display areas are updated )
  
      POLKeyUI 		( display, read, and evaluate )
      ABND		( abandon the IStk environment )
    ;
  ERRTRAP 		( if error, then restore saved user interface
			and error )
    POLResUI&Err
  POLRestoreUI		( restore the saved user interface )

  ClrDAsOK		( insure display areas are updated )
;

************************************************************
**
** Name: IStkDisp - The IStk display object
**
** ( --> )
**
************************************************************

NULLNAME IStkDisp		( --> )
::
  ( note the order of display areas 1 and 2 is switched to improve
  perceived display update performance )
  DA2OK? NOT_IT		( if the interactive stack area is invalid... )
    ?DispIStk		( ...then maybe display the interactive stack )
  DA1OK?NOTIT		( if the status area is invalid... )
    ?DispStatus		( ...then maybe display the system status )
  DA3OK?NOTIT		( if the menu labels are invalid... )
    ?DispMenu		( ...then maybe display them )
  ClrDAsOK
;

************************************************************
**
** Name: ?DispIStk - Display interactive stack according to
**         the key-in-buffer state
**
** ( --> )
**
** If there is no key in the key buffer, then display the
** interactive stack.  Otherwise, delay display but signal
** that the display is invalid.
**
************************************************************

NULLNAME ?DispIStk            ( --> )
::
  KEYINBUFFER?          ( if there's a key in the key buffer... )
  case :: SetDA2aBad SetDA2bBad ;	( then signal stack is bad )
  ( otherwise display the stack )
  TOADISP			( insure the alpha display is used )
  ClrAbbrevStk			( clear signal to abbreviate stack
  				display )
  SelLev@ SelPos@ #+-1		( calculate the highest-numbered stack
  				level to display )
  FIVE ZERO_DO (DO)		( for each level: )
    ISTOP-INDEX			( calculate the display row )
    OVER OVER#- #1+DUP 		( calculate the level number )
    #3+ SelLevDelta@ #+ ?PICK 	( try to get the stack level object )
    DispLevel! DROP		( display the one-row object )
    ISTOP-INDEX SelPos@ #= IT 	( if this display row is the selected
				position... )
      ShowSel!			( ...then indicate the selected level )
  LOOP
  DROP
  SetDA2Valid			( signal interactive stack area valid )
;

************************************************************
**
** Name: IStkKeys - The IStk key handler
**
** ( #KeyCode #Plane  --> KeyDef TRUE )
** ( #KeyCode #Plane  --> FALSE )
**
************************************************************

NULLNAME IStkKeys		( see stack diagram above )
::
  OVER SEVEN #< case2drop FALSE	( Allow menykeys )
  kpNoShift #=casedrop
	::
	  kcUpArrow   ?CaseKeyDef IStkUp&Rep	( move pointer up )
	  kcDownArrow ?CaseKeyDef IStkDn&Rep	( move pointer down )
	  kcNegate    ?CaseKeyDef IStkEdit	( edit the selected level )
	  kcBackspace ?CaseKeyDef IStkDrop	( discard selected level )
	  kcEnter     ?CaseKeyDef IStackEnter	( quit interactive stack )
	  kcOn        ?CaseKeyDef IStackAttn	( quit interactive stack )
	  kcNextRow    #=casedrpfls	( do std Next key definition )
	  kcLeftShift  #=casedrpfls	( do std Left-shift key definition )
	  kcRightShift #=casedrpfls	( do std Right-shift key definition )
	  DROP 'DoBadKeyT		( otherwise, bad key )
	;
  kpLeftShift #=casedrop
	::
	  kcUpArrow   ?CaseKeyDef IStkPgUp	( move pointer one page up )
	  kcDownArrow ?CaseKeyDef IStkPgDn	( move pointer one page down )
	  kcNegate    ?CaseKeyDef IStkEdit	( edit the selected level )
	  kcNextRow    #=casedrpfls	( do std Prev key definition )
	  kcLeftShift  #=casedrpfls	( do std Left-shift key definition )
	  kcRightShift #=casedrpfls	( do std Right-shift key definition )
	  DROP 'DoBadKeyT		( otherwise, bad key )
	;
  kpRightShift #=casedrop
	::
	  kcOn #=casedrpfls		( do std "off" key definition )
	  kcUpArrow   ?CaseKeyDef IStkFarUp	( move pointer far up )
	  kcDownArrow ?CaseKeyDef IStkFarDn	( move pointer far down )
	  kcNegate    ?CaseKeyDef IStkVisit	( visit the selected level )
	  kcNextRow    #=casedrpfls	( do std "1st Row" key definition )
	  kcLeftShift  #=casedrpfls	( do std Left-shift key definition )
	  kcRightShift #=casedrpfls	( do std Right-shift key definition )
	  DROP 'DoBadKeyT		( otherwise, bad key )
	;
  2DROP 'DoBadKeyT		( otherwise, bad key )
;

************************************************************
**
** Name: IStkUp&Rep - Move interactive stack one level up
**         and repeat until key is released
**
** ( --> )
**
************************************************************

NULLNAME IStkUp&Rep		( --> )
::
  TakeOver
  SelLev@ IStackDepth #1- #=case
    ( if the highest-numbered level is already selected... )
    DoBadKey			( ...then beep )
  TOADISP			( insure alpha display is used )
  REPEATERCH kcUpArrow
  ::
    SelLev@ IStackDepth #3- #=
      ( if the highest-numbered level is already selected... )
      ?SEMI			( ...then do nothing )
    UnShowSel!			( clear the current selected level 
				indication )
    SelLev@ #1+ SelLev!		( adjust the selected level )
    SelPos@ ONE #=ITE		( if the selected position is the
				top of the window... )
    ::				( ...then move the display down and
    				display a new level )
      CODE
	GOSBVL	=SAVPTR
	GOSBVL	=D0->Row1
	LC(5)	7*8*34
	A=A+C	A
	D1=A
	LC(5)	1*8*34
	A=A-C	A
	D0=A
	LC(5)	4*8*34
	GOSBVL	=MOVEUP
	GOVLNG	=GETPTRLOOP
      ENDCODE
      ONE SelLev@ DUP #4+ SelLevDelta@ #+ ?PICK DispLevel! DROP 
	( display the new selected level )
    ;
    :: 				( ...else move the selected position )
      SelPos@ #1- SelPos!	( adjust the selected position )
    ;
    ShowSel!			( indicate the new selected level )
    SG_ITE NOP SLOW
  ;
  SetDAsNoCh			( signal display areas are valid
				if they started out valid )
;

************************************************************
**
** Name: IStkDn&Rep - Move interactive stack one level down
**         and repeate until key is released
**
** ( --> )
**
************************************************************

NULLNAME IStkDn&Rep		( --> )
::
  TakeOver
  SelLev@ ONE #=case 		( if the lowest-numbered level is
				already selected... )
    DoBadKey			( ...then beep )
  TOADISP			( insure alpha display is used )
  REPEATERCH kcDownArrow
  ::
    SelLev@ #1=			( if the lowest-numbered level is
	  			already selected... )
      ?SEMI			( ...then do nothing )
    UnShowSel!			( clear the current selected level 
				indication )
    SelLev@ #1- SelLev!		( adjust the selected level )
    SelPos@ FIVE #=ITE		( if the selected position is the
				bottom of the window... )
    ::				( ...then move the display up and
    				display a new level )
      CODE
	GOSBVL	=SAVPTR
	GOSBVL	=D0->Row1
	LC(5)	2*8*34
	A=A+C	A
	D1=A
	LC(5)	8*34
	A=A+C	A
	D0=A
	LC(5)	4*8*34
	GOSBVL	=MOVEDOWN
	GOVLNG	=GETPTRLOOP
      ENDCODE
      FIVE SelLev@ DUP #4+ SelLevDelta@ #+ ?PICK DispLevel! DROP 
	( display the new selected level )
    ;
    :: 				( ...else move the selected position )
      SelPos@ #1+ SelPos!	( adjust the selected position )
    ;
    ShowSel!			( indicate the new selected level )
    SG_ITE NOP SLOW
  ;
  SetDAsNoCh			( signal display areas are valid
				if they started out valid )
;


************************************************************
**
** Name: IStkPgUp - Move interactive stack four levels up
**
** ( --> )
**
************************************************************

NULLNAME IStkPgUp		( --> )
::
  TakeOver
  IStackDepth SelLev@ 2DUP#=
    ( if the highest level is already selected... )
    case2drop DoBadKey		( ...then beep )
  2DUPSWAP FIVE #MAX #5- #>ITE 
    ( if the new level will be too high... )
    ::				( ...then adjust )
      DROP SelLev!		( drop the proposed level number and
      				select the highest level )
    ;
    ::				( ...else select the proposed level )
      #5+ SelLev!		( select the proposed level )
      DROP			( drop the highest level number )
    ;  
  ?AdjSelPos!			( maybe adjust the selected position )
  SetDA13NoCh			( signal that display areas 1 and 3
				have not changed )
;

************************************************************
**
** Name: IStkPgDn - Move interactive stack four levels down
**
** ( --> )
**
************************************************************

NULLNAME IStkPgDn		( --> )
::
  TakeOver
  SelLev@ DUP #1=		( if the lowest level is already
  				selected... )
    casedrop DoBadKey		( ...then beep )
  DUP SIX #<ITE			( if the new level will be too low... )
    ::				( ...then adjust )
      DROPONE SelLev!		( select the first level )
      FIVE SelPos!		( select the bottom position )
    ;
    ::				( ...else select the proposed level )
      #5- SelLev!		( select the proposed level )
      ?AdjSelPos!		( maybe adjust the selected position )
    ;
  SetDA13NoCh			( signal that display areas 1 and 3
				have not changed )
;
      
************************************************************
**
** Name: IStkFarUp - Move interactive stack to the highest-
**         numbered level
**
** ( --> )
**
************************************************************

NULLNAME IStkFarUp		( --> )
::
  TakeOver
  IStackDepth SelLev!		( select the highest level )
  ?AdjSelPos!			( maybe adjust the selected position )
  SetDA13NoCh			( signal that display areas 1 and 3
				have not changed )
;

************************************************************
**
** Name: IStkFarDn - Move interactive stack to level 1
**
** ( --> )
**
************************************************************

NULLNAME IStkFarDn		( --> )
::
  TakeOver
  ONE SelLev!			( set the selected level )
  FIVE SelPos!			( set the selected position )
  SetDA13NoCh			( signal that display areas 1 and 3
				are not changed )
;

************************************************************
**
** Name: IStkMenu - The non-saving, non-editing IStk menu
**
** ( --> IStkMenu )
**
************************************************************

NULLNAME IStkMenu		( --> IStkMenu )
::
  NoExitAction			( insure this menu is not saved in
				LastMenu )
  {
    { "ECHO" IStackEcho }
    { "VIEW"  
      { IStkView
	DoBadKey
	IStkIView
      }
    }
    { xROLL    IStackRoll }
    { xROLLD   IStackRollD }
    { "KEEP" 
	::
	  TakeOver 
	  DEPTH UNCOERCE CKNNOLASTWD DROP	( claim the entire stack )
	  SelLev@ 			( create the to-keep mata-ob )
	  DEPTH #1- OVER#- OVER#2+UNROLL
					( create the to-drop meta-ob )
	  psh				( swap the meta-obs )
	  N+1DROP ( "NDROP DROP" )	( drop the to-drop meta-ob and the 
					to-keep count )
	  ?AdjSel/Exit!			( maybe adjust the sel. level or exit )
	  SetDA13NoCh			( signal status and menu label areas
					unchanged )
	;
    }
    { "DRPN"   :: IStkDoN: NDROP ; }
* Menu page 2
    { xDUPN    IStackDupN }
    { xPICK    IStackPick }
    { "LEVEL"  IStackLevel }
    { "\8DLIST" :: IStkDoN: {}N ; }
    { "\8DPRG"  :: IStkDoN: ::N ; }
    { :: TakeOver "SRPL" hpdis? Std/BoxLabel ;
      :: TakeOver
         hpdisflag DUP TestUserFlag ITE ClrUserFlag SetUserFlag
         SetDA1NoCh
      ;
    }
  }
;

NULLNAME IStkDoN:
:: SelLev@&Ck 'REVAL ?AdjSel/Exit! SetDA13NoCh ;

************************************************************
**
** Name: IStkMenuEd - The editing IStk menu
**
** ( --> IStkMenuEd )
**
************************************************************

NULLNAME IStkMenuEd		( --> IStkMenuEd )
{
  { "ECHO" IStackEcho }
  { :: TakeOver "SRPL" hpdis? Std/BoxLabel ;
    :: TakeOver
       hpdisflag DUP TestUserFlag ITE ClrUserFlag SetUserFlag
       SetDA1NoCh
    ;
  }
}

************************************************************
**
** Name: IStkView - View/edit the selected level
**
** ( --> )
**
************************************************************

NULLNAME IStkView		( --> )
::
  TakeOver
  SelLev@ UNCOERCE
  Roll&View		( view/edit the specified level )
  ?AdjSel/Exit!		( maybe adjust the selected level or exit )
;

************************************************************
**
** Name: IStkIView - Indirect View/edit the selected level
**
** ( --> )
**
************************************************************

NULLNAME IStkIView		( --> )
::
  TakeOver
  SelLev@ PICK		( copy the selected object )
  '
  ::			( indirect-view the specified level )
    CK1NOLASTWD
    CK&DISPATCH1
    ONE Roll&View
    SIX Rcl&View
    SEVEN Rcl&View
  ;
  DoKeyOb
  ?AdjSel/Exit!		( maybe adjust the selected level or exit )
;

************************************************************
**
** Name: IStkEdit - Edit the selected level
**
** ( --> )
**
** Since this word is bound to a hard key, it is necessary
** to check for a nonexistent edit line before proceeding.
**
************************************************************

NULLNAME IStkEdit		( --> )
::
  TakeOver
  EditLExists? caseDoBadKey	( just beep if there is an edit
				line existent )
  SelLev@ UNCOERCE
  Roll&Edit		( edit the specified level )
  ?AdjSel/Exit!		( maybe adjust the selected level or exit )
;

************************************************************
**
** Name: IStkVisit - Visit the selected level
**
** ( --> )
**
** Since this word is bound to a hard key, it is necessary
** to check for a nonexistent edit line before proceeding.
**
************************************************************

NULLNAME IStkVisit		( --> )
::
  TakeOver
  EditLExists? caseDoBadKey	( just beep if there is an edit
				line existent )
  SelLev@ PICK 		( copy the selected object )
  ::			( vist the specified level )
    CK&DISPATCH1
    ONE Roll&Edit
    SIX Rcl&Edit
    SEVEN Rcl&Edit
  ;
  ?AdjSel/Exit!		( maybe adjust the selected level or exit )
;


************************************************************
**
** Name: IStkDrop - Discard the selected stack level
**
** ( obn ... obm+1 obm obm-1 ... ob1 
**   --> obn ... obm+1 obm-1 ... ob1 )
**
************************************************************

NULLNAME IStkDrop		( see above )
::
  TakeOver
  EditLExists? caseDoBadKey	( just beep if there is an edit
				line existent )
  SelLev@&Ck ROLLDROP 		( discard the selected object )
  ?AdjSel/Exit!		( maybe adjust the selected level or exit )
  SetDA13NoCh			( signal status and menu label areas
				unchanged )
;

************************************************************
**
** Name: ?AdjSel/Exit! - Maybe adjust the selected level and
**         display position or maybe exit according to the
**         change in the stack depth
**
** ( --> )
**
** To compact code, this word also signals a valid user stack.
**
************************************************************

NULLNAME ?AdjSel/Exit!		( --> )
::
  AtUserStack			( signal a valid user stack )
  DEPTH ZERO #=casedrop		( if the stack is empty... )
    SetISExit			( ...then signal to exit )
  DUP SelLev@ #<ITE		( if the stack depth < selected level... )
    SelLev!			( ...then select the highest level )
    DROP			( ...else drop the stack depth )
  ?AdjSelPos!			( maybe adjust the selected position )
;

************************************************************
**
** Name: ?AdjSelPos! - Maybe adjust the selected display 
**         position according to the current stack depth 
**         and selected level
**
** ( --> )
**
************************************************************

NULLNAME ?AdjSelPos!		( --> )
::
  IStackDepth 			( calculate interactive stack depth )
  SelLev@ #- FOUR #MIN #1+ 	( calculate maximum display position )

  SIX SelLev@ FIVE #MIN #- 	( calculate minimum display position )

  SelPos@ ROT #MIN #MAX SelPos!	( adjust with favor to minimum position
				and set new display position )
;

************************************************************
**
** IStk Utilities
**
************************************************************

NULLNAME ShowSel!		( --> )
::
ASSEMBLE
	CON(5)	=DOGROB
	REL(5)	+
	CON(5)	8
	CON(5)	6
	NIBGRB	110000
	NIBGRB	111000
	NIBGRB	111100
	NIBGRB	111110
	NIBGRB	111100
	NIBGRB	111000
	NIBGRB	110000
	NIBGRB	000000
+
RPL
 >SelPict!		( replace colon with pointer )
;

NULLNAME UnShowSel!		( --> )
::
ASSEMBLE
	CON(5)	=DOGROB
	REL(5)	+
	CON(5)	8
	CON(5)	6
	NIBGRB	000000
	NIBGRB	011000
	NIBGRB	011000
	NIBGRB	000000
	NIBGRB	011000
	NIBGRB	011000
	NIBGRB	000000
	NIBGRB	000000
+
RPL
  >SelPict!		( replace pointer with colon )
;

NULLNAME >SelPict!		( grobItem --> )
::
  ABUFF
  SelLev@ #>$ LEN$ #6* 		( calculate x-offset of pointer )
  SelPos@ #1+ #8*		( calulate y-offset of pointer )
  GROB!				( replace current item with new )
;

